@@ -1,6 +1,6 @@
 /*
 ** Zabbix
-** Copyright 2001-2022 Zabbix SIA
+** Copyright 2001-2024 Zabbix SIA
 **
 ** Licensed under the Apache License, Version 2.0 (the "License");
 ** you may not use this file except in compliance with the License.
@@ -19,7 +19,6 @@ package plugin
 
 import (
 	"context"
-	"crypto/tls"
 	"database/sql"
 	"fmt"
 	"net"
@@ -29,15 +28,33 @@ import (
 	"sync"
 	"time"
 
-	"git.zabbix.com/ap/plugin-support/tlsconfig"
-	"git.zabbix.com/ap/plugin-support/uri"
-	"git.zabbix.com/ap/plugin-support/zbxerr"
 	"github.com/jackc/pgx/v4/pgxpool"
 	"github.com/jackc/pgx/v4/stdlib"
 	"github.com/omeid/go-yarn"
+	"golang.zabbix.com/sdk/errs"
+	"golang.zabbix.com/sdk/log"
+	"golang.zabbix.com/sdk/metric"
+	"golang.zabbix.com/sdk/tlsconfig"
+	"golang.zabbix.com/sdk/uri"
 )
 
-const MinSupportedPGVersion = 100000
+const (
+	// pgx dns field names
+	password  = "password"
+	sslMode   = "sslmode"
+	rootCA    = "sslrootcert"
+	cert      = "sslcert"
+	key       = "sslkey"
+	cacheMode = "statement_cache_mode"
+
+	// connType
+	disable    = "disable"
+	require    = "require"
+	verifyCa   = "verify-ca"
+	verifyFull = "verify-full"
+
+	MinSupportedPGVersion = 100000
+)
 
 type PostgresClient interface {
 	Query(ctx context.Context, query string, args ...interface{}) (rows *sql.Rows, err error)
@@ -58,6 +75,11 @@ type PGConn struct {
 	address        string
 }
 
+type connID struct {
+	uri       uri.URI
+	cacheMode string
+}
+
 var errorQueryNotFound = "query %q not found"
 
 // Query wraps pgxpool.Query.
@@ -72,9 +94,10 @@ func (conn *PGConn) Query(ctx context.Context, query string, args ...interface{}
 }
 
 // QueryByName executes a query from queryStorage by its name and returns a single row.
-func (conn *PGConn) QueryByName(ctx context.Context, queryName string, args ...interface{}) (rows *sql.Rows, err error) {
-	if sql, ok := (*conn.queryStorage).Get(queryName + sqlExt); ok {
-		normalizedSQL := strings.TrimRight(strings.TrimSpace(sql), ";")
+func (conn *PGConn) QueryByName(ctx context.Context, queryName string, args ...any) (*sql.Rows, error) {
+	querySQL, ok := (*conn.queryStorage).Get(queryName + sqlExt)
+	if ok {
+		normalizedSQL := strings.TrimRight(strings.TrimSpace(querySQL), ";")
 
 		return conn.Query(ctx, normalizedSQL, args...)
 	}
@@ -94,9 +117,12 @@ func (conn *PGConn) QueryRow(ctx context.Context, query string, args ...interfac
 }
 
 // QueryRowByName executes a query from queryStorage by its name and returns a single row.
-func (conn *PGConn) QueryRowByName(ctx context.Context, queryName string, args ...interface{}) (row *sql.Row, err error) {
-	if sql, ok := (*conn.queryStorage).Get(queryName + sqlExt); ok {
-		normalizedSQL := strings.TrimRight(strings.TrimSpace(sql), ";")
+func (conn *PGConn) QueryRowByName(
+	ctx context.Context, queryName string, args ...any,
+) (*sql.Row, error) {
+	querySQL, ok := (*conn.queryStorage).Get(queryName + sqlExt)
+	if ok {
+		normalizedSQL := strings.TrimRight(strings.TrimSpace(querySQL), ";")
 
 		return conn.QueryRow(ctx, normalizedSQL, args...)
 	}
@@ -123,9 +149,8 @@ func (conn *PGConn) updateAccessTime() {
 
 // ConnManager is a thread-safe structure for manage connections.
 type ConnManager struct {
-	sync.Mutex
-	connMutex      sync.Mutex
-	connections    map[uri.URI]*PGConn
+	connectionsMu  sync.Mutex
+	connections    map[connID]*PGConn
 	keepAlive      time.Duration
 	connectTimeout time.Duration
 	callTimeout    time.Duration
@@ -135,11 +160,12 @@ type ConnManager struct {
 
 // NewConnManager initializes connManager structure and runs Go Routine that watches for unused connections.
 func NewConnManager(keepAlive, connectTimeout, callTimeout,
-	hkInterval time.Duration, queryStorage yarn.Yarn) *ConnManager {
+	hkInterval time.Duration, queryStorage yarn.Yarn,
+) *ConnManager {
 	ctx, cancel := context.WithCancel(context.Background())
 
 	connMgr := &ConnManager{
-		connections:    make(map[uri.URI]*PGConn),
+		connections:    make(map[connID]*PGConn),
 		keepAlive:      keepAlive,
 		connectTimeout: connectTimeout,
 		callTimeout:    callTimeout,
@@ -154,26 +180,26 @@ func NewConnManager(keepAlive, connectTimeout, callTimeout,
 
 // closeUnused closes each connection that has not been accessed at least within the keepalive interval.
 func (c *ConnManager) closeUnused() {
-	c.connMutex.Lock()
-	defer c.connMutex.Unlock()
+	c.connectionsMu.Lock()
+	defer c.connectionsMu.Unlock()
 
-	for uri, conn := range c.connections {
+	for ci, conn := range c.connections {
 		if time.Since(conn.lastTimeAccess) > c.keepAlive {
 			conn.client.Close()
-			delete(c.connections, uri)
-			Impl.Debugf("[%s] Closed unused connection: %s", Name, uri.Addr())
+			delete(c.connections, ci)
+			Impl.Debugf("[%s] Closed unused connection: %s", Name, ci.uri.Addr())
 		}
 	}
 }
 
 // closeAll closes all existed connections.
 func (c *ConnManager) closeAll() {
-	c.connMutex.Lock()
-	for uri, conn := range c.connections {
+	c.connectionsMu.Lock()
+	for ci, conn := range c.connections {
 		conn.client.Close()
-		delete(c.connections, uri)
+		delete(c.connections, ci)
 	}
-	c.connMutex.Unlock()
+	c.connectionsMu.Unlock()
 }
 
 // housekeeper repeatedly checks for unused connections and closes them.
@@ -194,22 +220,14 @@ func (c *ConnManager) housekeeper(ctx context.Context, interval time.Duration) {
 }
 
 // create creates a new connection with given credentials.
-func (c *ConnManager) create(uri uri.URI, details tlsconfig.Details) (*PGConn, error) {
-	c.connMutex.Lock()
-	defer c.connMutex.Unlock()
-
-	if _, ok := c.connections[uri]; ok {
-		// Should never happen.
-		panic("connection already exists")
-	}
-
+func (c *ConnManager) create(ci connID, details tlsconfig.Details) (*PGConn, error) {
 	ctx := context.Background()
 
-	host := uri.Host()
-	port := uri.Port()
+	host := ci.uri.Host()
+	port := ci.uri.Port()
 
-	if uri.Scheme() == "unix" {
-		socket := uri.Addr()
+	if ci.uri.Scheme() == "unix" {
+		socket := ci.uri.Addr()
 		host = filepath.Dir(socket)
 
 		ext := filepath.Ext(filepath.Base(socket))
@@ -220,48 +238,86 @@ func (c *ConnManager) create(uri uri.URI, details tlsconfig.Details) (*PGConn, e
 		port = ext[1:]
 	}
 
-	dbname, err := url.QueryUnescape(uri.GetParam("dbname"))
+	dbname, err := url.QueryUnescape(ci.uri.GetParam("dbname"))
 	if err != nil {
 		return nil, err
 	}
 
-	dsn := fmt.Sprintf("host=%s port=%s dbname=%s user=%s",
-		host, port, dbname, uri.User())
-
-	if uri.Password() != "" {
-		dsn += " password=" + uri.Password()
-	}
-
-	client, err := createTLSClient(dsn, c.connectTimeout, details)
+	client, err := createClient(
+		createDNS(
+			host,
+			port,
+			dbname,
+			ci.uri.User(),
+			ci.uri.Password(),
+			ci.cacheMode,
+			details,
+		),
+		c.connectTimeout,
+	)
 	if err != nil {
 		return nil, err
 	}
 
 	serverVersion, err := getPostgresVersion(ctx, client)
 	if err != nil {
+		client.Close()
 		return nil, err
 	}
 
 	if serverVersion < MinSupportedPGVersion {
+		client.Close()
 		return nil, fmt.Errorf("PostgreSQL version %d is not supported", serverVersion)
 	}
 
-	c.connections[uri] = &PGConn{
+	Impl.Debugf("[%s] Created new connection: %s", Name, ci.uri.Addr())
+
+	return &PGConn{
 		client:         client,
 		callTimeout:    c.callTimeout,
 		version:        serverVersion,
 		lastTimeAccess: time.Now(),
 		ctx:            ctx,
 		queryStorage:   &c.queryStorage,
-		address:        uri.Addr(),
-	}
-
-	Impl.Debugf("[%s] Created new connection: %s", Name, uri.Addr())
-
-	return c.connections[uri], nil
+		address:        ci.uri.Addr(),
+	}, nil
 }
 
-func createTLSClient(dsn string, timeout time.Duration, details tlsconfig.Details) (*sql.DB, error) {
+func createDNS(host, port, dbname, user, pass, mode string, details tlsconfig.Details) string {
+	dsn := fmt.Sprintf("host=%s port=%s dbname=%s user=%s", host, port, dbname, user)
+
+	tmp := map[string]string{
+		password:  pass,
+		sslMode:   details.TlsConnect,
+		rootCA:    details.TlsCaFile,
+		cert:      details.TlsCertFile,
+		key:       details.TlsKeyFile,
+		cacheMode: mode,
+	}
+
+	for k, v := range tmp {
+		if v != "" {
+			dsn = fmt.Sprintf("%s %s=%s", dsn, k, v)
+		}
+	}
+
+	return dsn
+}
+
+func renameTLS(in string) string {
+	switch in {
+	case "required":
+		return "require"
+	case "verify_ca":
+		return "verify-ca"
+	case "verify_full":
+		return "verify-full"
+	default:
+		return in
+	}
+}
+
+func createClient(dsn string, timeout time.Duration) (*sql.DB, error) {
 	config, err := pgxpool.ParseConfig(dsn)
 	if err != nil {
 		return nil, err
@@ -277,54 +333,104 @@ func createTLSClient(dsn string, timeout time.Duration, details tlsconfig.Detail
 		return conn, err
 	}
 
-	config.ConnConfig.TLSConfig, err = getTLSConfig(details)
+	return stdlib.OpenDB(*config.ConnConfig), nil
+}
+
+// GetConnection returns an existing connection or creates a new one.
+func (c *ConnManager) GetConnection(
+	ci connID, params map[string]string, //nolint:gocritic
+) (*PGConn, error) {
+	conn := c.getConn(ci)
+	if conn != nil {
+		return conn, nil
+	}
+
+	details, err := getTlsDetails(params)
 	if err != nil {
 		return nil, err
 	}
 
-	return stdlib.OpenDB(*config.ConnConfig), nil
-}
-
-func getTLSConfig(details tlsconfig.Details) (*tls.Config, error) {
-	switch details.TlsConnect {
-	case "required":
-		return &tls.Config{InsecureSkipVerify: true}, nil
-	case "verify_ca":
-		return tlsconfig.CreateConfig(details, true)
-	case "verify_full":
-		return tlsconfig.CreateConfig(details, false)
-	}
-
-	return nil, nil
-}
-
-// get returns a connection with given uri if it exists and also updates lastTimeAccess, otherwise returns nil.
-func (c *ConnManager) get(uri uri.URI) *PGConn {
-	c.connMutex.Lock()
-	defer c.connMutex.Unlock()
-
-	if conn, ok := c.connections[uri]; ok {
-		conn.updateAccessTime()
-		return conn
-	}
-
-	return nil
-}
-
-// GetConnection returns an existing connection or creates a new one.
-func (c *ConnManager) GetConnection(uri uri.URI, details tlsconfig.Details) (conn *PGConn, err error) {
-	c.Lock()
-	defer c.Unlock()
-
-	conn = c.get(uri)
-
-	if conn == nil {
-		conn, err = c.create(uri, details)
-	}
-
+	conn, err = c.create(ci, details)
 	if err != nil {
-		err = zbxerr.ErrorConnectionFailed.Wrap(err)
+		return nil, errs.Wrap(err, "failed to create connection")
 	}
 
-	return
+	return c.setConn(ci, conn), nil
+}
+
+// get returns a connection with given uri if it exists and also updates
+// lastTimeAccess, otherwise returns nil.
+func (c *ConnManager) getConn(cd connID) *PGConn { //nolint:gocritic
+	c.connectionsMu.Lock()
+	defer c.connectionsMu.Unlock()
+
+	conn, ok := c.connections[cd]
+	if !ok {
+		return nil
+	}
+
+	conn.updateAccessTime()
+
+	return conn
+}
+
+func (c *ConnManager) setConn(cd connID, conn *PGConn) *PGConn { //nolint:gocritic
+	c.connectionsMu.Lock()
+	defer c.connectionsMu.Unlock()
+
+	existingConn, ok := c.connections[cd]
+	if ok {
+		conn.client.Close() //nolint:errcheck,gosec
+
+		log.Debugf("Closed redundant connection: %s", cd.uri.Addr())
+
+		return existingConn
+	}
+
+	c.connections[cd] = conn
+
+	return conn
+}
+
+func getTlsDetails(params map[string]string) (tlsconfig.Details, error) {
+	tlsType := renameTLS(params[tlsConnectParam])
+	validateCA := true
+
+	if tlsType == "" {
+		tlsType = disable
+	}
+
+	details := tlsconfig.NewDetails(
+		params[metric.SessionParam],
+		tlsType,
+		params[tlsCAParam],
+		params[tlsCertParam],
+		params[tlsKeyParam],
+		params[uriParam],
+		disable,
+		require,
+		verifyCa,
+		verifyFull,
+	)
+
+	if tlsType == disable || tlsType == require {
+		validateCA = false
+	}
+
+	err := details.Validate(validateCA, false, false)
+	return details, err
+}
+
+func createConnID(params map[string]string) (connID, error) {
+	u, err := uri.NewWithCreds(
+		fmt.Sprintf("%s?dbname=%s", params[uriParam], url.QueryEscape(params[databaseParam])),
+		params[userParam],
+		params[passwordParam],
+		uriDefaults,
+	)
+	if err != nil {
+		return connID{}, err
+	}
+
+	return connID{uri: *u, cacheMode: params[cacheModeParam]}, nil
 }
